# Copyright (C) 2006, Polychronis Ypodimatopoulos
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import os, sys, sha, time
from shutil import move
from sets import Set
from string import upper
from math import ceil
from struct import *

from cerebro.common import *

class FileManager:
	def __init__(self, packet_size):
		# Dealer class to connect to
		self.packet_size	= packet_size
		
		# Holds a list of (hash, filename)
		self.filenames		= []
		
		# a long string that contains all available hashes provided for fast ADV compilation
		self.hashes			= ""
		
		try:
			os.makedirs(FILEDIR)
		except:
			print FILEDIR, "exists already, good."
		#try:
			#os.makedirs(TEMPDIR)
		#except:
			##if DEBUG > 1:
			#print TEMPDIR, "exists already, good."

		self.scan_local_files()
		
		i = 0
		for f in self.filenames:
			print i, f[0], f[1].strip()
			i = i + 1
	
	def scan_local_files(self):
#		out( "Scanning for files in "+ FILEDIR )
		forbiddend		= []
		forbiddenf		= []

#		forbiddend.append("MedleyMoinGettingStarted")
#		forbiddend.append("MedleyNetworkInfo")
#		forbiddend.append("MedleyOct06DemoFaq")
#		forbiddend.append("PeerPresenceLog")
#		forbiddend.append("cache")
#		forbiddenf.append("edit-lock")
#		forbiddenf.append("edit-log")

		self.hashes		= ""
		self.filenames	= []
		directories = [ FILEDIR ]
		while len( directories ) > 0:
			directory = directories.pop()
			traverse = 1
			for fdir in forbiddend:
				if directory.find(fdir) >= 0:
#					print "forbiden dir", directory, fdir
					traverse = 0
			if traverse:
		#		print "ok", directory
				for name in os.listdir( directory ):
					fullpath = os.path.join( directory, name )
					if os.path.isfile( fullpath ):
						for ffil in forbiddenf:
							if fullpath.find(ffil) >= 0:
								traverse = 0
						if traverse:
							f = file(fullpath, 'r+')
							contents = f.read()
							f.close()
							
							newhash = sha.new( contents )
							newhash = newhash.digest()
							hexhash = self.ntohash( newhash )
							self.add_hash( hexhash, name )
							
							#for md in os.popen("sha1sum "+fullpath):
								#d(fhash, fname) = md.split("  ", 1)
								#fhash = upper( fhash )
#								print fname.strip()
								#self.add_hash( fhash, fname.strip() )
					elif os.path.isdir( fullpath ):
						directories.append( fullpath )

	def add_hash(self, fhash, fname):
		#print "--",fname
		self.filenames.append( (fhash, fname) )
		self.hashes += self.hashton( fhash ) # append hash in 20-byte representation

	
	# Given a file name, remove the hash and filename from lists
	def remove_hash(self, fname):
		for (fhash, filename) in self.filenames:
			if fname == filename:
				self.filenames.remove( (fhash, filename) )
				fhash = self.hashton( fhash )
				print "BEFORE:", len(self.hashes)
				pos = self.hashes.find( fhash )
				if pos > -1:
					self.hashes = self.hashes[:pos] + self.hashes[pos+20:]
				print "AFTER :", len(self.hashes)

	
	def get_filename(self, fhash):
		for (fh, fn) in self.filenames:
			if fhash == fh: return fn.strip()
		return ""
	
	# Pack a 40-byte readable hash into its 20-byte equivalent
	def hashton(self, fhash):
		if len(fhash) != 40:
			print "Error: hashton: wrong input size: %d" % len(fhash)
			return ""
		res = ''.join([chr( int(fhash[b:b+2], 16) ) for b in xrange(0,40,2)])
		return res
		
	# unpack a 20-byte hash into a 40-byte readable format
	def ntohash(self, bytes):
		if len(bytes) != 20:
			print "Error: ntohash: wrong input size: %d" % len(bytes)
			return ""
		res = ''.join(["%02X" % ord(b) for b in bytes])
		return res


	def save_file(self, fhash, data):
		filename = fhash + ".dat"
		file = open(FILEDIR + filename, 'wb+')
		file.write( data )
		file.close()
		self.add_hash( fhash, filename )
		
		
	def handle_request(self, payload):
		raw_hash = payload[:20]
		fhash, pieces = self.ntohash( raw_hash ), payload[20:]
			
		fname = self.get_filename(fhash)
		if not fname:
			print "FM: returning nothing. fhash:", fhash, fname
			return ("", "", "")

		#fname = FILEDIR + fname
		print "FM: I was requested to send", fname
		try:
			f = file(FILEDIR + fname, 'r+') # Try to get exclusive access
		except:
			print "Error: I could not open file:", fname
			return ""
		
		#while len(fname) < FILENAME_SIZE: # Pad with spaces in order to reach FILENAME_SIZE
			#fname += " "
		
		# If the rest of the request is empty, then the client asked for the whole file
		if len(pieces) == 0:
			contents = f.read()
			f.close()
			print "Filemanager will return with:", len(fhash), len(contents), "bytes"
			return raw_hash, fname, contents
		else:
			print "FM: Panic: The request contained data I did not expect:", pieces
			sys.exit()

		
